unit Ship;

(*********************************************
TShip->TPolygonSprite

A good example of a TSprite that applies its
own custom Move method.  In this case, the Move
method handles acceleration, deceleration,
momentum and thrust.

Theoretically, this class isn't bound to the
asteroids app, since it doesn't use the main
form of the app.
*********************************************)

interface

uses

  SysUtils, WinTypes, WinProcs, Classes, Graphics, Controls, WinGPnt, Sprite,
  LkupTabl, Polygon, Utility;

const

  MAX_SPEED = 200;

type

  TExplosionBit = record
     pt: TPoint;
     x, y: integer;
  end;

  TShip = class( TPolygonSprite )
  private
     bWillDie: boolean;
     FThrust: boolean;
     bFired: boolean;
     dx, dy: single;
     nCountX, nCountY: integer;
     nInnerX, nInnerY: integer;
     nAccelerate, nDecelerate: integer;
     nMod: byte;
     FHyper: byte;
     bExploding: boolean;
     FInv: boolean;
     FInvCount: byte;
     arExplosion: array[1..50] of TExplosionBit;
  protected
     function GetCanFire: boolean;
     procedure AddVertices( const nCol: byte ); override;
     function GetInHyper: boolean;
     procedure SetInv( b: boolean );
  public
     bDead: boolean;
     nSpeedX, nSpeedY: integer;
     procedure DestroyShip( var nLives: integer ); { Doesn't really destroy the sprite, just makes it look like it does! }
     procedure Fire;
     procedure Hyperspace;
     procedure StopFire;
     procedure Move; override;
     procedure Render( winGPnt: TWinGPaintBox ); override;
     property InHyperSpace: boolean read GetInHyper;
     property Invulnerable: boolean read FInv write SetInv;
     property Thrusting: boolean read FThrust write FThrust;
     property CanFire: boolean read GetCanFire;
  end;

implementation

uses
  Missile;

(*********************************************
This procedure must be overriden for all
sprites descended from TPolygonSprite
*********************************************)
procedure TShip.AddVertices( const nCol: byte );
begin
  lstVertices.Add( TVertex.Create( Point( 0, -9 ), nCol ) );
  lstVertices.Add( TVertex.Create( Point( 7, 9 ), nCol ) );
  lstVertices.Add( TVertex.Create( Point( 0, 5 ), nCol ) ); { Thrust position! }
  lstVertices.Add( TVertex.Create( Point( -7, 9 ), nCol ) );
end;

function TShip.GetCanFire: boolean;
begin
  Result := not bFired;
end;

(*********************************************
This move method augments the inherited method,
and handles inertia, momentum and thrust.  We
still call inherited Move, which handles alot of
useful stuff, because we're not setting any of
the standard move variables of the sprite.  In
effect, the bulk of inerited Move never executes
because of this fact.
*********************************************)
procedure TShip.Move;

  procedure AdjustPos( var nPos, nSpeed, nInner: integer );
  var
     nAbs: integer;
     nAdd, nTime: integer;
  begin
     nAbs := Abs( nSpeed );
     if nAbs > 100 then
        begin
           nAdd := ( ( nAbs - 100 ) div 10 );
           if nAdd < 2 then
              nAdd := 2;
           if nSpeed > 0 then
              Inc( nPos, nAdd )
           else
              Dec( nPos, nAdd );
        end
     else
        begin
           Inc( nInner );
           Inc( nMod );
           nAdd := 1;
           case nAbs of
              50..100:
                 begin
                    nTime := 1;
                    if nMod mod 2 > 0 then
                       nAdd := 2
                    else
                       nAdd := 1;
                 end;
              30..49:
                 nTime := 1;
              20..29:
                 nTime := 2;
              10..19:
                 nTime := 3;
              5..9:
                 nTime := 4;
           else
              nTime := 5;
           end;
           if nInner >= nTime then
              begin
                 nInner := 0;
                 if nSpeed > 0 then
                    Inc( nPos, nAdd )
                 else
                    Dec( nPos, nAdd );
              end;
        end;
  end;

  procedure SlowDown( var n: integer );
  begin
     if n = 0 then
        Exit;
     if n > 0 then
        Dec( n, ( n div 20 ) + 1 )
     else
        Inc( n, Abs( n Div 20 ) + 1 );
  end;

begin
  inherited Move;

  if FInvCount > 0 then
     Dec( FInvCount );
  if FInvCount = 0 then
     FInv := FALSE;

  if FHyper > 0 then
     begin
        Dec( FHyper );
        if FHyper = 0 then
           begin
              if bExploding then
                 Invulnerable := TRUE;
              bExploding := FALSE;
           end;
        Exit;
     end;

  if bWillDie then
     begin
        bDead := TRUE;
        Exit;
     end;

{ Increace speed vectors appropriately when thrusting }
  Inc( nAccelerate );
  if nAccelerate > 10 then
     begin
        nAccelerate := 0;
        if Thrusting then
           begin
              dx := 20 * lk.lkupCos[Angle];
              dy := 20 * lk.lkupSin[Angle];
              Inc( nSpeedX, Trunc( dx + 0.5 ) );
              Dec( nSpeedY, Trunc( dy + 0.5 ) );
           end;
     end;

{ Cap speed at maximum }
  if nSpeedX > MAX_SPEED then
     nSpeedX := MAX_SPEED;
  if nSpeedY > MAX_SPEED then
     nSpeedY := MAX_SPEED;

  if nSpeedX < -MAX_SPEED then
     nSpeedX := -MAX_SPEED;
  if nSpeedY < -MAX_SPEED then
     nSpeedY := -MAX_SPEED;

{ Test the current speed and adjust ship position if appropriate }
  if nSpeedX <> 0 then
     AdjustPos( ptPosition.X, nSpeedX, nInnerX );
  if nSpeedY <> 0 then
     AdjustPos( ptPosition.Y, nSpeedY, nInnerY );

{ Slowly decelerate until stop if thrust not applied }
  if Thrusting then
     Exit;

  Inc( nDecelerate );
  if nDecelerate = 10 then
     begin
        nDecelerate := 0;
        SlowDown( nSpeedX );
        SlowDown( nSpeedY );
     end;
end;

procedure TShip.Fire;
var
  ptDest: TPoint;
begin
  if not bFired then
     begin
        bFired := TRUE;
        with TMissile.CreateMissile( 50, 64 + ( ( Tag - 1 ) * 20 ) ) do
           begin
{ Project the angle out to a destination point }
              ptDest.X := self.ptPosition.X + Trunc( 100 * lk.lkupCos[Angle] );
              ptDest.Y := self.ptPosition.Y - Trunc( 100 * lk.lkupSin[Angle] );
              Position := self.ptPosition;
              Destination := ptDest;
              Tag := self.Tag;
              RegisterSprite( w );
           end;
     end;
end;

(*********************************************
Intended to be called when the Return key is
released.
*********************************************)
procedure TShip.StopFire;
begin
  bFired := FALSE;
end;

(*********************************************
Draw a little thrust near the exhaust vertex.
(I know this could be done better, but I didn't
want to spend alot of time on it!
*********************************************)
procedure TShip.Render( winGPnt: TWinGPaintBox );
var
  pt, pt_: TPoint;
  i: integer;
begin
  if FHyper > 0 then
     begin
        if bExploding then
           begin
              for i := 1 to 50 do
                 begin
                    with arExplosion[i] do
                       begin
                          if Random( 255 ) > Abs( x ) then
                             if x < 0 then
                                Dec( pt.X )
                             else
                                Inc( pt.X );
                          if Random( 255 ) > Abs( y ) then
                             if y < 0 then
                                Dec( pt.Y )
                             else
                                Inc( pt.Y );
                          if ( pt.X >= 0 ) and ( pt.X < w.PhysicalWidth ) and
                             ( pt.Y >= 0 ) and ( pt.Y < w.PhysicalHeight ) then
                             w.Pixels[pt.X, pt.Y] := Random( 16 ) + 64;
                       end;
                 end;
           end;
        Exit;
     end;

  if bDead then
     Exit;

  if FInv then
     if ( FInvCount div 5 ) mod 2 > 0 then
        Exit;

  inherited Render( winGPnt );
  if Thrusting then
     begin
        pt := TVertex( lstVertices[2] ).ptPoint;
        Inc( pt.X, ptPosition.X );
        Inc( pt.Y, ptPosition.Y );
        if ( pt.X > 4 ) and ( pt.X + 4 < winGPnt.PhysicalWidth ) and
           ( pt.Y > 4 ) and ( pt.Y + 4 < winGPnt.PhysicalHeight ) then
           begin
              for i := 1 to 12 do
                 begin
                    pt_ := pt;
                    pt_.X := pt_.X + Random( 4 ) - 2 + Random( 4 ) - 2;
                    pt_.Y := pt_.Y + Random( 4 ) - 2 + Random( 4 ) - 2;
                    winGPnt.Pixels[pt_.X, pt_.Y] := Random( 16 ) + 64;
                 end;
           end;
     end;
end;

function TShip.GetInHyper: boolean;
begin
  Result := FHyper > 0;
end;

procedure TShip.Hyperspace;
begin
  FHyper := 75;
  nSpeedX := 0;
  nSpeedY := 0;
  Angle := Random( 360 );
  ptPosition.X := Random( w.PhysicalWidth );
  ptPosition.Y := Random( w.PhysicalHeight );
end;

procedure TShip.DestroyShip( var nLives: integer );
var
  i: integer;
begin
  Dec( nLives );
  bWillDie := nLives = 0;
  FHyper := 255; { Hyperspace flag used for destruction also ... since it has similar effect }
  nSpeedX := 0;
  nSpeedY := 0;
  Angle := 0;
  bExploding := TRUE;
{ Position the explosion bits, and set their X/Y speeds }
  for i := 1 to 50 do
     begin
        with arExplosion[i] do
           begin
              pt.X := ptPosition.X;
              pt.Y := ptPosition.Y;
              x := Random( 255 );
              if Random( 100 ) > 50 then
                 x := -x;
              y := Random( 255 );
              if Random( 100 ) > 50 then
                 y := -y;
           end;
     end;
  ptPosition.Y := w.PhysicalHeight div 2;
  case Tag of
     1:
        ptPosition.X := w.PhysicalWidth div 2 - 100;
     2:
        ptPosition.X := w.PhysicalWidth div 2 + 100;
  end;
end;

procedure TShip.SetInv( b: boolean );
begin
  FInv := b;
  if b then
     FInvCount := 255;
end;

end.
